/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/utsname.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <tiuser.h>
#include <netconfig.h>
#include <netdir.h>

#define HOSTFILE		"/etc/net/testnet/hosts"
#define SERVICEFILE	"/etc/net/testnet/services"
#define TADDRLEN		6	/* total address length */
#define NADDRLEN		4	/* network address length */
#define OCTMAX		255	/* maximum value of an octet */

struct host {
	struct host	*h_next;
	char		*h_name;
	ulong_t		h_naddr;
};

struct svc {
	struct svc	*s_next;
	char		*s_name;
	ushort_t		s_port;
};

struct addr {
	ulong_t		a_n;
	ushort_t		a_p;
};

static struct host *hosts;
static struct svc *svcs;

static void clrhosts(void);
static void clrsvcs(void);
static int rdhosts(void);
static int rdsvcs(void);
static int counthosts(char *);
static struct svc *findsvc(char *);
static struct host *findhost(struct host *, char *);
static struct host *findaddr(struct host *, struct addr *);

extern int _nderror;
struct nd_addrlist *
_netdir_getbyname(struct netconfig *ncp,
	struct nd_hostserv *hsp)
{
	char *hostname;
	struct host *hp;
	struct svc *sp;
	int i, nhosts;
	struct utsname self;
	struct netbuf *np;
	struct nd_addrlist *nalp;
	struct addr *ap;

	_nderror = ND_OK;

	/*
	 * Check for special aliases we don't support.
	 */
	if ((strcmp(hsp->h_host, HOST_ANY) == 0) ||
	  (strcmp(hsp->h_host, HOST_BROADCAST) == 0)) {
		_nderror = ND_NOHOST;
		return(NULL);
	}

	/*
	 * If local alias is specified, call uname(2)
	 * to get the local host name.  Otherwise use
	 * name supplied by caller.
	 */
	if (strcmp(hsp->h_host, HOST_SELF) == 0) {
		if (uname(&self) < 0) {
			_nderror = ND_SYSTEM;
			return(NULL);
		}
		hostname = self.nodename;
	} else {
		hostname = hsp->h_host;
	}

	/*
	 * Read the hosts database, if we haven't already
	 * done so.
	 */
	if ((hosts == NULL) && (rdhosts() < 0)) {
		_nderror = ND_SYSTEM;
		return(NULL);
	}

	/*
	 * Read the services database, if we haven't already
	 * done so.
	 */
	if ((svcs == NULL) && (rdsvcs() < 0)) {
		_nderror = ND_SYSTEM;
		return(NULL);
	}

	/*
	 * Count the number of host entries matching the
	 * given host name.
	 */
	if ((nhosts = counthosts(hostname)) == 0) {
		_nderror = ND_NOHOST;
		return(NULL);
	}

	/*
	 * See if the requested service is defined.
	 */
	if ((sp = findsvc(hsp->h_serv)) == NULL) {
		_nderror = ND_NOSERV;
		return(NULL);
	}

	/*
	 * Allocate the address list to be returned.
	 */
	if ((nalp = malloc(sizeof(struct nd_addrlist))) ==
	  NULL) {
		_nderror = ND_NOMEM;
		return(NULL);
	}
	nalp->n_cnt = nhosts;
	nalp->n_addrs = malloc(nhosts*sizeof(struct netbuf));
	if (nalp->n_addrs == NULL) {
		free(nalp);
		_nderror = ND_NOMEM;
		return(NULL);
	}
	hp = hosts;
	np = (struct netbuf *)nalp->n_addrs;
	for (i = 0; i < nalp->n_cnt; i++, np++) {

		/*
		 * Allocate a netbuf address buffer for each
		 * address.
		 */
		if ((np->buf = malloc(TADDRLEN)) == NULL) {
			np = (struct netbuf *)nalp->n_addrs;
			for (; i > 0; i--, np++)
				free(np->buf);
			free(nalp->n_addrs);
			free(nalp);
			_nderror = ND_NOMEM;
			return(NULL);
		}
		np->len = TADDRLEN;
		np->maxlen = TADDRLEN;

		/*
		 * Find the host and copy the address into
		 * the netbuf structure.
		 */
		hp = findhost(hp, hostname);
		ap = (struct addr *)np->buf;
		ap->a_n = hp->h_naddr;
		ap->a_p = sp->s_port;
	}
	return(nalp);
}

struct nd_hostservlist *
_netdir_getbyaddr(struct netconfig *ncp, struct netbuf *np)
{
	struct host *hp;
	struct svc *sp;
	int i, nhosts;
	struct nd_hostservlist *hslp;
	struct nd_hostserv *hsp;
	struct addr *ap;

	_nderror = ND_OK;

	/*
	 * Make sure the address is the proper length.
	 */
	if (np->len != TADDRLEN) {
		_nderror = ND_BADARG;
		return(NULL);
	}

	/*
	 * Read the hosts database, if we haven't already
	 * done so.
	 */
	if ((hosts == NULL) && (rdhosts() < 0)) {
		_nderror = ND_SYSTEM;
		return(NULL);
	}

	/*
	 * Read the services database, if we haven't already
	 * done so.
	 */
	if ((svcs == NULL) && (rdsvcs() < 0)) {
		_nderror = ND_SYSTEM;
		return(NULL);
	}

	/*
	 * Check if the port number is assigned to a
	 * service.
	 */
	ap = (struct addr *)np->buf;
	for (sp = svcs; sp; sp = sp->s_next)
		if (sp->s_port == ap->a_p)
			break;
	if (sp == NULL) {
		_nderror = ND_NOSERV;
		return(NULL);
	}

	/*
	 * Check for matching hosts based on the network
	 * portion of the address.
	 */
	nhosts = 0;
	for (hp = hosts; hp; hp = hp->h_next)
		if (hp->h_naddr == ap->a_n)
			nhosts++;
	if (nhosts == 0) {
		_nderror = ND_NOHOST;
		return(NULL);
	}

	/*
	 * Allocate memory for the list of hosts and
	 * services that match the given address.
	 */
	hslp = malloc(sizeof(struct nd_hostservlist));
	if (hslp == NULL) {
		_nderror = ND_NOMEM;
		return(NULL);
	}
	hsp = malloc(sizeof(struct nd_hostserv) * nhosts);
	if (hsp == NULL) {
		free(hslp);
		_nderror = ND_NOMEM;
		return(NULL);
	}
	hslp->h_cnt = nhosts;
	hslp->h_hostservs = hsp;
	hp = hosts;
	for (i = 0; i < nhosts; i++, hsp++) {
		/*
		 * Copy the name of the matching service.
		 */
		hsp->h_serv = strdup(sp->s_name);
		if (hsp->h_serv == NULL) {
			hsp = hslp->h_hostservs;
			for (; i > 0; i--, hsp++) {
				free(hsp->h_host);
				free(hsp->h_serv);
			}
			free(hslp->h_hostservs);
			free(hslp);
			_nderror = ND_NOMEM;
			return(NULL);
		}

		/*
		 * Find the next matching host and copy
		 * its name.
		 */
		hp = findaddr(hp, ap);
		hsp->h_host = strdup(hp->h_name);
		if (hsp->h_host == NULL) {
			free(hsp->h_serv);
			hsp = hslp->h_hostservs;
			for (; i > 0; i--, hsp++) {
				free(hsp->h_host);
				free(hsp->h_serv);
			}
			free(hslp->h_hostservs);
			free(hslp);
			_nderror = ND_NOMEM;
			return(NULL);
		}
	}
	return(hslp);
}

int
_netdir_options(struct netconfig *ncp, int option, int fd,
  void *arg)
{
	int ret;
	struct nd_mergearg *nmp;

	_nderror = ND_OK;
	ret = 0;
	switch (option) {
	case ND_MERGEADDR:
		/*
		 * No local aliases supported.  Just return
		 * a copy of the local universal address.
		 */
		nmp = (struct nd_mergearg *)arg;
		nmp->m_uaddr = strdup(nmp->s_uaddr);
		if (nmp->m_uaddr == NULL) {
			_nderror = ND_NOMEM;
			ret = -1;
		}
		break;

	default:
		/*
		 * No other commands supported.
		 */
		_nderror = ND_NOCTRL;
		ret = -1;
	}
	return(ret);
}

char *
_taddr2uaddr(struct netconfig *ncp, struct netbuf *np)
{
	char *p;
	struct addr *ap;
	char buf[128];

	_nderror = ND_OK;

	/*
	 * Make sure the address length is valid.
	 */
	if (np->len != TADDRLEN) {
		_nderror = ND_BADARG;
		return(NULL);
	}
	ap = (struct addr *)np->buf;

	/*
	 * Convert the address to ASCII and return a copy.
	 */
	sprintf(buf, "%d.%d.%d.%d.%d.%d", (ap->a_n>>24),
	  ((ap->a_n>>16)&0xff), ((ap->a_n>>8)&0xff),
	  (ap->a_n&0xff), ((ap->a_p>>8)&0xff), (ap->a_p&0xff));
	if ((p = strdup(buf)) == NULL)
		_nderror = ND_NOMEM;
	return(p);
}

struct netbuf *
_uaddr2taddr(struct netconfig *ncp, char *uaddr)
{
	uint_t i1, i2, i3, i4, i5, i6;
	struct netbuf *np;
	struct addr *ap;

	_nderror = ND_OK;

	/*
	 * Parse the universal address.
	 */
	if (sscanf(uaddr, "%d.%d.%d.%d.%d.%d", &i1, &i2, &i3,
	  &i4, &i5, &i6) != 6) {
		_nderror = ND_BADARG;
		return(NULL);
	}

	/*
	 * Validate each octet.
	 */
	if ((i1 > OCTMAX) || (i2 > OCTMAX) || (i3 > OCTMAX) ||
	  (i4 > OCTMAX) || (i5 > OCTMAX) || (i6 > OCTMAX)) {
		_nderror = ND_BADARG;
		return(NULL);
	}

	/*
	 * Allocate a netbuf structure with a buffer large
	 * enough to hold a transport address.
	 */
	if ((np = malloc(sizeof(struct netbuf))) == NULL) {
		_nderror = ND_NOMEM;
		return(NULL);
	}
	if ((np->buf = malloc(TADDRLEN)) == NULL) {
		free(np);
		_nderror = ND_NOMEM;
		return(NULL);
	}

	/*
	 * Initialize the netbuf structure with the binary
	 * address.
	 */
	np->len = TADDRLEN;
	np->maxlen = TADDRLEN;
	ap = (struct addr *)np->buf;
	ap->a_n = (i1<<24)|(i2<<16)|(i3<<8)|i4;
	ap->a_p = (i5<<8)|i6;
	return(np);
}

/*
 * Free all host information.
 */
static void
clrhosts(void)
{
	struct host *hp, *nhp;

	hp = hosts;
	while (hp != NULL) {
		nhp = hp->h_next;
		free(hp);
		hp = nhp;
	}
	hosts = NULL;
}

/*
 * Free all service information.
 */
static void
clrsvcs(void)
{
	struct svc *sp, *nsp;

	sp = svcs;
	while (sp != NULL) {
		nsp = sp->s_next;
		free(sp);
		sp = nsp;
	}
	svcs = NULL;
}

static int
rdhosts(void)
{
	FILE *fp;
	uint_t i1, i2, i3, i4;
	struct host *hp;
	char hname[256];
	char buf[256];

	/*
	 * Open the host database.
	 */
	if ((fp = fopen(HOSTFILE, "r")) == NULL)
		return(-1);

	/*
	 * Check one line at a time.
	 */
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		/*
		 * Skip comment lines.
		 */
		if (buf[0] == '#')
			continue;

		/*
		 * Skip lines not in the format of a host name
		 * followed by the network address in universal
		 * format.
		 */
		if (sscanf(buf, "%s %d.%d.%d.%d", hname, &i1, &i2,
		  &i3, &i4) != 5)
			continue;

		/*
		 * Validate the octet values.
		 */
		if ((i1 > OCTMAX) || (i2 > OCTMAX) ||
		  (i3 > OCTMAX) || (i4 > OCTMAX))
			continue;

		/*
		 * Allocate a host structure.  If the allocation
		 * fails, free the entire host list.
		 */
		hp = malloc(sizeof(struct host)+strlen(hname)+1);
		if (hp == NULL) {
			fclose(fp);
			clrhosts();
			return(-1);
		}

		/*
		 * Initialize the host structure and put it on
		 * the front of the list.
		 */
		hp->h_name = (char *)hp + sizeof(struct host);
		strcpy(hp->h_name, hname);
		hp->h_naddr = (i1<<24)|(i2<<16)|(i3<<8)|i4;
		hp->h_next = hosts;
		hosts = hp;
	}

	/*
	 * Don't forget to close the file.
	 */
	fclose(fp);
	return(0);
}

static int
rdsvcs(void)
{
	FILE *fp;
	uint_t port;
	struct svc *sp;
	char sname[256];
	char buf[256];

	/*
	 * Open the service database.
	 */
	if ((fp = fopen(SERVICEFILE, "r")) == NULL)
		return(-1);

	/*
	 * Check one line at a time.
	 */
	while (fgets(buf, sizeof(buf), fp) != NULL) {
		/*
		 * Skip comment lines.
		 */
		if (buf[0] == '#')
			continue;

		/*
		 * Skip lines not in the format of a service
		 * name followed by a port number.
		 */
		if (sscanf(buf, "%s %d", sname, &port) != 2)
			continue;

		/*
		 * Validate the port number.
		 */
		if (port > USHRT_MAX)
			continue;

		/*
		 * Allocate a service structure.  If the
		 * allocation fails, free the entire
		 * service list.
		 */
		sp = malloc(sizeof(struct svc)+strlen(sname)+1);
		if (sp == NULL) {
			fclose(fp);
			clrsvcs();
			return(-1);
		}

		/*
		 * Initialize the service structure and put it
		 * on the front of the list.
		 */
		sp->s_name = (char *)sp + sizeof(struct svc);
		strcpy(sp->s_name, sname);
		sp->s_port = (ushort_t)port;
		sp->s_next = svcs;
		svcs = sp;
	}

	/*
	 * Don't forget to close the file.
	 */
	fclose(fp);
	return(0);
}

static int
counthosts(char *name)
{
	struct host *hp;
	int n = 0;

	for (hp = hosts; hp; hp = hp->h_next)
		if (strcmp(hp->h_name, name) == 0)
			n++;
	return(n);
}

static struct svc *
findsvc(char *name)
{
	struct svc *sp;

	for (sp = svcs; sp; sp = sp->s_next)
		if (strcmp(sp->s_name, name) == 0)
			return(sp);
	return(NULL);
}

static struct host *
findhost(struct host *hp, char *name)
{
	if (hp != hosts)
		hp = hp->h_next;
	while (hp != NULL) {
		if (strcmp(hp->h_name, name) == 0)
			return(hp);
		hp = hp->h_next;
	}
	return(NULL);
}

static struct host *
findaddr(struct host *hp, struct addr *ap)
{
	if (hp != hosts)
		hp = hp->h_next;
	while (hp != NULL) {
		if (hp->h_naddr == ap->a_n)
			return(hp);
		hp = hp->h_next;
	}
	return(NULL);
}
