/*
 * Copyright (c) 1993 Addison-Wesley Publishing Company.
 * All rights reserved.  This source code is provided
 * for its instructional value, and is not guaranteed
 * for any particular purpose.
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>

void timeout(int);
int sameaddr(struct sockaddr *, int, struct sockaddr *,
	int);

extern void fatal(const char *, ...);

void
rwho(int argc, char *argv[], int family,
	struct sockaddr *ap, uint_t alen)
{
	int fd, i, rlen;
	struct sigaction sa;
	char raddr[128];
	char buf[1024];

	if (argc < 2) {
		fprintf(stderr, "usage: rwho host [options]\n");
		exit(1);
	}

	/*
	 * Create a socket endpoint.
	 */
	if ((fd = socket(family, SOCK_DGRAM, 0)) < 0)
		fatal("rwho: cannot create socket");

	/*
	 * Build the command string to send to the server.
	 */
	strcpy(buf, "who");
	for (i = 2; i < argc; i++) {
		strcat(buf, " ");
		strcat(buf, argv[i]);
	}

	/*
	 * Set an alarm so we don't block forever.
	 */
	sa.sa_handler = timeout;
	sa.sa_flags = 0;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGALRM, &sa, NULL);
	alarm(20);

	/*
	 * Send the request.
	 */
	if (sendto(fd, buf, strlen(buf)+1, 0, ap, alen) < 0)
		fatal("rwho: cannot send request");

	/*
	 * Receive response messages until we encounter an
	 * error or receive the last message (a single byte
	 * containing 0).
	 */
	for ( ; ; ) {
		rlen = sizeof(raddr);
		if ((i = recvfrom(fd, buf, sizeof(buf), 0,
		  (struct sockaddr *)raddr, &rlen)) < 0)
			fatal("rwho: cannot receive data");

		/*
		 * Only process the message if it is from the
		 * server we contacted.
		 */
		if (sameaddr(ap, alen, (struct sockaddr *)raddr,
		  rlen)) {
			/*
			 * If this is the last message, exit.
			 */
			if ((i == 1) && (buf[0] == '\0'))
				exit(0);
			/*
			 * Write the response on the standard output.
			 */
			write(1, buf, i);
		}
	}
}

void
timeout(int sig)
{
	fprintf(stderr, "rwho: command timed out\n");
	exit(1);
}

/*
 * Return 0 if the two addresses are different, and
 * 1 if they are the same.
 */
int
sameaddr(struct sockaddr *a1p, int l1, struct sockaddr *a2p,
	int l2)
{
	int i;

	/*
	 * If the lengths or address families differ,
	 * then the addresses don't match.
	 */
	if ((l1 != l2) || (a1p->sa_family != a2p->sa_family))
		return(0);

	/*
	 * We already checked the address family.  Now
	 * compare the data portion.
	 */
	l1 -= sizeof(a1p->sa_family);
	for (i = 0; i < l1; i++)
		if (a1p->sa_data[i] != a2p->sa_data[i])
			return(0);
	return(1);
}
